#pragma once

#include <atomic>
#include <thread>
#include <iostream>
#include <sstream>
#include <utility>
#include "folly/FBVector.h"
#include "folly/MPMCQueue.h"
#include "folly/LifoSem.h"
#include "task.h"
#include "relation.h"
#include "profiling.h"
#include "validation-result.h"

#ifdef _PROFILE
#include <iomanip>
#endif

extern folly::LifoSem flushReadySemaphore;
extern std::atomic<uint64_t> flushCounter;

namespace Validator
{
extern const int THREAD_COUNT;

struct ValidationSubtask
{
    ValidationQueries *v;
    Query *q;

    ValidationSubtask()
    { }

    ValidationSubtask(ValidationQueries *_v, Query *_q) 
    : v(_v), q(_q)
    { }
};

struct ThreadContext
{
    ThreadContext();

    void mainLoop();

    void validation(ValidationSubtask &subtask);

    folly::fbvector<uint32_t> columnHashedValue;
    size_t columnHashedValueCompute;

    folly::fbvector<uint64_t> resultSet;

    const Query::Column *columnsBegin;
    const Query::Column *columnsEqualEnd;
    const Query::Column *columnsEnd;

    uint64_t *bitColumnIntersection;
    uint64_t bitColumnIntersectionReservedRows;
    uint64_t firstUnknownValidationId;
    folly::fbvector<ValidationResult> validationResults;

#ifdef _PROFILE
    TimestampStack timeStack;
    ProfilingStats timeValidation;
    ProfilingStats timeIntersection;
    ProfilingStats timeExtraction;
    ProfilingStats timeSmallValidation;
    ProfilingStats timeBloomFilterValidation;
    ProfilingStats *timeRelation;

    std::string profilingStatsToString()
    {
        std::ostringstream oss;
        oss << "Validation: " << timeValidation.toString() << "\n> Small: " << timeSmallValidation.toString() << "\n> Bloom Intersection: " << timeIntersection.toString() << "\n> Bloom Extraction: " << timeExtraction.toString() << "\n> Bloom Validation: " << timeBloomFilterValidation.toString() << "\nRelations: ";
        for (int i = 0;i < relationsEnd - relations;i++)
            oss << std::fixed << std::setw(9) << std::setiosflags(std::ios_base::left) << std::setprecision(6) << timeRelation[i].totalTimeToSeconds();
        return oss.str();
    }
#endif
};
}

extern folly::MPMCQueue<folly::fbvector<Validator::ValidationSubtask>> validationQueue;
